/*
    BFilter - a smart ad-filtering web proxy
    Copyright (C) 2002-2003  Joseph Artsimovich <joseph_a@mail.ru>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifndef LEXGEN_H_
#define LEXGEN_H_

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <iostream>
#include <string>
#include <vector>
#include <list>
#include <map>
#include <set>
#include "nfa.h"
#include "nfamove.h"

using namespace std;

class LexGen
{
public:
	typedef enum { OPT_LAZY = 1 } options_type;
	LexGen() : m_rule_id_generator(0) {}
	void includeCode(const string& code) { m_include_code = code; }
	void useSubclass(const string& cname) { m_use_subclass = cname; }
	void addRule(int scond, const NFA& nfa, const string& action, const NFA* trailer=0, int options=0);
	void generateLexer(ostream& out);
private:
	class Rule
	{
	public:
		Rule(int id, const NFA& nfa, const string& action, int scond, const NFA* trailer, int options);
		Rule(const Rule& rule);
		~Rule();
		Rule& operator=(const Rule& rule);
		int m_id;
		const NFA* m_nfa;
		string m_action;
		int m_scond;
		const NFA* m_trailer;
		int m_options;
		int m_main_length;
		int m_trailer_length;
	};
	class NStateInfo
	{
	public:
		NStateInfo(int rule, bool lazy) : m_rule(rule), m_is_lazy(lazy),
			m_is_trailer(false), m_is_quickfinish(false), m_is_semifinish(false), m_is_longfinish(false) {}
		int m_rule;
		bool m_is_lazy;
		bool m_is_trailer;
		bool m_is_quickfinish;
		bool m_is_semifinish;
		bool m_is_longfinish;
		// only one of the finishes may be set
	};
	class DStateInfo
	{
	public:
	DStateInfo() : m_nMainStates(0), m_nTrailerStates(0) {}
		int m_nMainStates;
		int m_nTrailerStates;
	};
	class LexerRuleComposition
	{
	public:
		typedef map<int,int,less<int> > finishes_type;
		LexerRuleComposition(const list<Rule>& rules);
		int nStates() const { return m_nstates; }
		list<NFAMove> getMovesFrom(int state) const;
		NStateInfo getStateInfo(int state) const;
	private:
		int newState(int rule, bool lazy);
		int m_nstates;
		vector<list<NFAMove> > m_moves;
		list<NFAMove> m_start_moves;
		typedef map<int, NStateInfo, less<int> > stateinfo_type;
		stateinfo_type m_stateinfo;
	};
	class DState
	{
	public:
		DState(const LexerRuleComposition* composition);
		DState(const DState& dstate);
		~DState() { delete[] m_moves; }
		DState& operator=(const DState& dstate);
		bool operator==(const DState& dstate) const { return m_stateset==dstate.m_stateset; }
		bool operator!=(const DState& dstate) const { return m_stateset!=dstate.m_stateset; }
		const LexerRuleComposition* m_composition;
		set<int> m_stateset;
		int m_state_num;
		int m_finish_for;
		bool m_is_quickfinish;
		bool m_is_lazy_finish;
		set<int> m_semifinish_for;
		bool m_dead_end;
		short *m_moves;
	};
	typedef map<int, Rule, less<int> > rules_by_id_type;
	typedef map<int, list<Rule>, less<int> > rules_by_scond_type;
	int buildGotoTree(vector<string>& rule_code, int from, int to);
	DState eClosureInitial(const LexerRuleComposition& nfa);
	DState eClosure(const DState& dstate);
	DState move(DState& dstate, char ch);
	rules_by_id_type m_rules_by_id;
	rules_by_scond_type m_rules_by_scond;
	int m_rule_id_generator;
	string m_include_code;
	string m_use_subclass;
};

#endif
