/*
  tre.h - TRE internal definitions

  Copyright (C) 2001, 2002 Ville Laurikari <vl@iki.fi>.

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License version 2 (June
  1991) as published by the Free Software Foundation.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/

#ifdef TRE_DEBUG
#include <stdio.h>
#define DPRINT(msg) do {printf msg; fflush(stdout);} while(0)
#else /* !TRE_DEBUG */
#define DPRINT(msg) do { } while(0)
#endif /* !TRE_DEBUG */

#if defined HAVE_MBRTOWC
#define tre_mbrtowc(pwc, s, n, ps) (mbrtowc((pwc), (s), (n), (ps)))
#elif defined HAVE_MBTOWC
#define tre_mbrtowc(pwc, s, n, ps) (mbtowc((pwc), (s), (n)))
#endif

/* Define the character types and functions. */
#ifdef TRE_WCHAR
/* Wide characters. */
typedef wchar_t tre_char_t;
typedef wint_t tre_cint_t;
typedef wctype_t tre_ctype_t;
#define TRE_CHAR_MAX WCHAR_MAX
#ifdef TRE_MULTIBYTE
#define TRE_MB_CUR_MAX MB_CUR_MAX
#else /* !TRE_MULTIBYTE */
#define TRE_MB_CUR_MAX 1
#endif /* !TRE_MULTIBYTE */
#define tre_isctype iswctype
#define tre_isalnum iswalnum
#define tre_islower iswlower
#define tre_isupper iswupper
#define tre_tolower towlower
#define tre_toupper towupper
#define tre_ctype   wctype
#define tre_strlen  wcslen
#else /* !TRE_WCHAR */
/* 8 bit characters. */
typedef unsigned char tre_char_t;
typedef short tre_cint_t;
typedef int (*tre_ctype_t)(tre_cint_t);
#define TRE_CHAR_MAX 255
#define TRE_MB_CUR_MAX 1
#define tre_isctype(c, type) ( (type)(c) )
int tre_isalnum(tre_cint_t c);
int tre_islower(tre_cint_t c);
int tre_isupper(tre_cint_t c);
#define tre_tolower tolower
#define tre_toupper toupper
tre_ctype_t tre_ctype(const char *name);
#define tre_strlen  strlen
#endif /* !TRE_WCHAR */

#define REG_NOTAGS (REG_NOTEOL << 1)
typedef enum { STR_WIDE, STR_BYTE, STR_MBS } tre_str_type_t;

/* TNFA transition definition.  Each transition has a range of accepted
   characters, pointer to the next state and the ID number of that state,
   a -1 -terminated array of tags, assertion bitmap, and character class
   assertions.  A TNFA state is an array of transitions, the terminator is
   a transition with state == NULL: */
typedef struct tnfa_transition tre_tnfa_transition_t;

struct tnfa_transition {
  tre_cint_t code_min;
  tre_cint_t code_max;
  tre_tnfa_transition_t *state;
  int state_id;
  int *tags;
  int assertions;
  tre_ctype_t class;
  tre_ctype_t *neg_classes;
};


/* Assertions. */
#define ASSERT_AT_BOL             1          /* Beginning of line. */
#define ASSERT_AT_EOL             2          /* End of line. */
#define ASSERT_CHAR_CLASS         4          /* Character class. */
#define ASSERT_CHAR_CLASS_NEG     8          /* Character classes, negated. */
#define ASSERT_AT_BOW            16          /* Beginning of word. */
#define ASSERT_AT_EOW            32          /* End of word. */
#define ASSERT_AT_WB             64          /* Word boundary. */
#define ASSERT_AT_WB_NEG        128          /* Word boundary, negated. */
#define ASSERT_LAST             128

/* Tag directions. */
#define TAG_MINIMIZE 0
#define TAG_MAXIMIZE 1


/* Instructions to compute submatch register values after a successful
   match.  */
struct submatch_data {
  /* The start offset of this submatch can be computed by taking the tag
     specified by `so_tag' and subtracting `so_offset'. */
  int so_offset;
  int so_tag;
  /* The end offset of this submatch can be computed by taking the tag
     specified by `eo_tag' and subtracting `eo_offset'. */
  int eo_offset;
  int eo_tag;
  /* This is a tag that gives a left boundary for the submatch start.  If
     the computed value is smaller than the boundary tag, then the submatch
     is from a previous repetition and must be disregarded. */
  int boundary;
  /* marker_tag > 0: If the value of the tag specified by `marker_tag' is -1
     then this subexpression does not participate in the match.
     marker_tag = 0: This subexpression always participates in the match.  */
  int marker_tag;
  /* `marker_boundary' > 0: If the value of the tag specified by `marker_tag'
     is smaller than the value of the tag specified by `marker_boundary'
     then this subexpression does not participate in the match, even if
     the marker tag is positive. */
  int marker_boundary;
};

typedef struct submatch_data tre_submatch_data_t;




/* TNFA definition. */
typedef struct tnfa tre_tnfa_t;

struct tnfa {
  tre_tnfa_transition_t *transitions;
  int num_transitions;
  tre_tnfa_transition_t *initial;
  tre_tnfa_transition_t *final;
  tre_submatch_data_t *submatch_data;
  char *firstpos_chars;
  int first_char;
  int num_submatches;
  int *tag_directions;
  int *marker_offs;
  int num_tags;
  int end_tag;
  int num_states;
  int cflags;
};
