/*
    BFilter - a smart ad-filtering web proxy
    Copyright (C) 2002-2003  Joseph Artsimovich <joseph_a@mail.ru>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "bfilter.h"
#include <string>
#include <sstream>
#include <algorithm>
#include <map>
#include "conf.h"
#include "httperrormessages.h"
#include "network.h"
#include "tselect.h"
#include "syscall.h"
#include "childfinishexception.h"
#include "debug.h"

#ifdef WIN32

#include <winsock2.h>

#else

#include <errno.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <unistd.h>

#endif

#define SERVER_READ_TIMEOUT  500
#define CLIENT_READ_TIMEOUT  100
#define SERVER_WRITE_TIMEOUT 100
#define CLIENT_WRITE_TIMEOUT 100
#define IDLE_TIMEOUT         300

using namespace std;

BFilter::BFilter(Network::Socket csock)
:	m_csock(csock), m_ssock(Network::INVALID_SOCK), m_html_processor(this),
	m_server_port(-1), m_request_stream(*this), m_response_stream(*this)
{}

BFilter::~BFilter()
{
	list<HttpClientRequest *>::iterator iter = m_requests_to_be_served.begin();
	for (; iter != m_requests_to_be_served.end(); ++iter) {
		delete *iter;
	}
}

void
BFilter::run()
{
	fd_set rfds, wfds;
	Network::Socket maxfd;
	TSelect tselect;
	struct timeval sread_timeout;
	sread_timeout.tv_sec = SERVER_READ_TIMEOUT;
	sread_timeout.tv_usec = 0;
	struct timeval cread_timeout;
	cread_timeout.tv_sec = CLIENT_READ_TIMEOUT;
	cread_timeout.tv_usec = 0;
	struct timeval swrite_timeout;
	swrite_timeout.tv_sec = SERVER_WRITE_TIMEOUT;
	swrite_timeout.tv_usec = 0;
	struct timeval cwrite_timeout;
	cwrite_timeout.tv_sec = CLIENT_WRITE_TIMEOUT;
	cwrite_timeout.tv_usec = 0;
	struct timeval idle_timeout;
	idle_timeout.tv_sec = IDLE_TIMEOUT;
	idle_timeout.tv_usec = 0;
	int sread_timer = 0;
	int cread_timer = 0;
	int swrite_timer = 0;
	int cwrite_timer = 0;
	int idle_timer = 0;
	char recv_buf[8192];
	while (true) {
		FD_ZERO(&rfds);
		FD_ZERO(&wfds);
		FD_SET(m_csock, &rfds);
		maxfd = m_csock;
		if (requestStream().isInitialState()) {
			if (cread_timer) {
				tselect.removeTimer(cread_timer);
				cread_timer = 0;
			}
		} else {
			if (!cread_timer)
				cread_timer = tselect.addTimer(cread_timeout);
		}
		if (requestStream().isInitialState() && responseStream().isInitialState()) {
			if (!idle_timer)
				idle_timer = tselect.addTimer(idle_timeout);
		} else if (idle_timer) {
			tselect.removeTimer(idle_timer);
			idle_timer = 0;
		}
		if (!m_data_for_client.empty()) {
			FD_SET(m_csock, &wfds);
			if (!cwrite_timer)
				cwrite_timer = tselect.addTimer(cwrite_timeout);
		}
		if (!m_data_for_server.empty()) {
			if (!Network::isValidSocket(m_ssock)) {
				makeServerConnection();
			}
			FD_SET(m_ssock, &wfds);
			if (m_ssock > maxfd)
				maxfd = m_ssock;
			if (!swrite_timer)
				swrite_timer = tselect.addTimer(swrite_timeout);
		}
		if (Network::isValidSocket(m_ssock)) {
			FD_SET(m_ssock, &rfds);
			if (m_ssock > maxfd)
				maxfd = m_ssock;
			if (!sread_timer && !responseStream().isInitialState())
				sread_timer = tselect.addTimer(sread_timeout);
		} else if (sread_timer) {
			tselect.removeTimer(sread_timer);
			sread_timer = 0;
		}
		int res = tselect.select(maxfd+1, &rfds, &wfds, NULL);
		if (res < 0) {
			// error
			DEBUGLOG("select() error");
			finishSession();
		} else if (res == 0) {
			set<int> timers = tselect.getTriggeredTimers();
			if (timers.find(cread_timer) != timers.end() ||
			    timers.find(cwrite_timer) != timers.end() ||
			    timers.find(idle_timer) != timers.end()) {
				finishSession();
			} else {
				closeServerConnection(false);
				responseStream().resetState();
			}
			break;
		}
		if (FD_ISSET(m_csock, &rfds)) {
			if (cread_timer) {
				tselect.removeTimer(cread_timer);
				cread_timer = 0;
			}
			if (idle_timer) {
				tselect.removeTimer(idle_timer);
				idle_timer = 0;
			}
			int len = -1;
			SYSCALL(len = recv(m_csock, recv_buf, sizeof(recv_buf), 0));
			if (len < 0) {
				// error
				DEBUGLOG("Error reading from client");
				finishSession();
			} else if (len == 0) {
				// client eof
				DEBUGLOG("client EOF");
				requestStream().consumeEOF();
				finishSession();
			}
			string msg(recv_buf, recv_buf+len);
			LOG_TRAFFIC_FROM_CLIENT(msg);
			requestStream().consumeDataChunk(msg);
		}
		if (Network::isValidSocket(m_ssock) && FD_ISSET(m_ssock, &rfds)) {
			if (sread_timer) {
				tselect.removeTimer(sread_timer);
				sread_timer = 0;
			}
			if (idle_timer) {
				tselect.removeTimer(idle_timer);
				idle_timer = 0;
			}
			int len = -1;
			SYSCALL(len = recv(m_ssock, recv_buf, sizeof(recv_buf), 0));
			if (len < 0) {
				// error
				DEBUGLOG("Error reading from server: " << Network::getLastError());
				responseStream().consumeEOF();
				closeServerConnection(true);
				responseStream().resetState();
			} else if (len == 0) {
				// server eof
				DEBUGLOG("Server closes connection");
				responseStream().consumeEOF();
				closeServerConnection(true);
			} else {
				string msg(recv_buf, recv_buf+len);
				LOG_TRAFFIC_FROM_SERVER(msg);
				responseStream().consumeDataChunk(msg);
			}
		}
		if (FD_ISSET(m_csock, &wfds) && !m_data_for_client.empty()) {
			if (cwrite_timer) {
				tselect.removeTimer(cwrite_timer);
				cwrite_timer = 0;
			}
			int len = -1;
			SYSCALL(len = send(m_csock, m_data_for_client.data(), m_data_for_client.length(), 0));
			if (len <= 0) {
				DEBUGLOG("Error writing to client");
				finishSession();
			}
			LOG_TRAFFIC_TO_CLIENT(m_data_for_client.substr(0, len));
			m_data_for_client.erase(0, len);
		}
		if (Network::isValidSocket(m_ssock) && FD_ISSET(m_ssock, &wfds) && !m_data_for_server.empty()) {
			if (swrite_timer) {
				tselect.removeTimer(swrite_timer);
				swrite_timer = 0;
			}
			int len = -1;
			SYSCALL(len = send(m_ssock, m_data_for_server.c_str(), m_data_for_server.length(), 0));
			if (len <= 0) {
				DEBUGLOG("Error writing to server");
				if (responseStream().isForwardingInProgress()) {
					finishSession();
				} else {
					closeServerConnection(true);
					responseStream().resetState();
				}
			}
			LOG_TRAFFIC_TO_SERVER(m_data_for_server.substr(0, len));
			m_data_for_server.erase(0, len);
		}
	}
	if (Network::isValidSocket(m_ssock)) {
		Network::closeSocket(m_ssock);
		m_ssock = Network::INVALID_SOCK;
	}
	// m_csock will be closed by the creator of the object
}

void
BFilter::dataForClient(const string& data)
{
	m_data_for_client += data;
}

void
BFilter::dataForServer(const string& data)
{
	m_data_for_server += data;
}

void
BFilter::finishSession()
{
	DEBUGLOG("Closing connections and finishing session");
	flushClientData();
	if (Network::isValidSocket(m_ssock)) {
		Network::closeSocket(m_ssock);
		m_ssock = Network::INVALID_SOCK;
	}
	// we dont close m_csock, because it will be closed by the creator of this object
	throw ChildFinishException();
}

void
BFilter::flushClientData()
{
	fd_set rfds, wfds;
	TSelect tselect;
	struct timeval tv;
	char buf[2048];
	tv.tv_sec = CLIENT_WRITE_TIMEOUT;
	tv.tv_usec = 0;
	int tid = tselect.addTimer(tv);
	while (!m_data_for_client.empty()) {
		FD_ZERO(&rfds);
		FD_ZERO(&wfds);
		FD_SET(m_csock, &rfds);
		FD_SET(m_csock, &wfds);
		int res = tselect.select(m_csock+1, &rfds, &wfds, NULL);
		if (res <= 0) {
			// timeout or error
			break;
		}
		if (FD_ISSET(m_csock, &rfds)) {
			int len = -1;
			SYSCALL(len = recv(m_csock, buf, sizeof(buf), 0));
			if (len <= 0) {
				// eof or error
				break;
			}
		}
		if (FD_ISSET(m_csock, &wfds)) {
			int len = -1;
			SYSCALL(len = send(m_csock, m_data_for_client.data(), m_data_for_client.length(), 0));
			if (len <= 0) {
				break;
			}
			LOG_TRAFFIC_TO_CLIENT(m_data_for_client.substr(0, len));
			m_data_for_client.erase(0, len);
			tselect.resetTimer(tid);
		}
	}
	DEBUGLOG("flushClientData() finishes");
}

void
BFilter::switchToServer(const string& host, int port)
{
	DEBUGLOG("switchToServer(" << host << ":" << port << ")");
#ifdef DEBUG
	if (!requestsSentToServer().empty()) {
		DEBUGLOG("WARNING: pending_requests not empty in BFilter::switchToServer()");
	}
#endif
	if (m_server_host == host && m_server_port == port) {
		return;
	}
	if (Network::isValidSocket(m_ssock)) {
		Network::closeSocket(m_ssock);
		m_ssock = Network::INVALID_SOCK;
	}
	m_server_host = host;
	m_server_port = port;
}

void
BFilter::makeServerConnection()
{
	DEBUGLOG("makeServerConnection(" << m_server_host << ":" << m_server_port << ")");
	if (Network::isValidSocket(m_ssock)) {
		Network::closeSocket(m_ssock);
		m_ssock = Network::INVALID_SOCK;
	}
	string host;
	int port;
	if (Conf::getInstance().usingProxy()) {
		host = Conf::getInstance().proxyHost();
		port = Conf::getInstance().proxyPort();
		DEBUGLOG("connecting to proxy: " << host << ":" << port);
	} else {
		host = m_server_host;
		port = m_server_port;
	}
	NetworkError err;
	m_ssock = Network::tcpConnect(host.c_str(), port, &err);
	if (Network::isValidSocket(m_ssock)) {
		return;
	} else {
		if (!responseStream().isForwardingInProgress()) {
			switch (err.getErrorCode()) {
			case NetworkError::ERR_UNKNOWN_HOST:
				dataForClient(HttpErrorMessages::errDNS(
					m_requests_sent_to_server.front()->getOrigRequestLine(), host));
				break;
			case NetworkError::ERR_CONNECTION_REFUSED:
				dataForClient(HttpErrorMessages::errConnRefused(
					m_requests_sent_to_server.front()->getOrigRequestLine(), host, port));
				break;
			default:
				dataForClient(HttpErrorMessages::errInternal(
					m_requests_sent_to_server.front()->getOrigRequestLine(), err.getErrorMessage()));
			}
		}
		finishSession();
	}
}

void
BFilter::closeServerConnection(bool server_has_closed_it)
{
	DEBUGLOG("closeServerConnection()");
	serverInfo().reset();
	if (!Network::isValidSocket(m_ssock)) {
		DEBUGLOG("connection already closed");
	} else {
		Network::closeSocket(m_ssock);
		m_ssock = Network::INVALID_SOCK;
		m_server_host.erase();
		m_server_port = 0;
		if (m_response_stream.isForwardingInProgress()) {
			DEBUGLOG("PANIC: closing server connection in the middle of response");
			finishSession();
		}
		if (server_has_closed_it) {
			if (serverInfo().getReceivedRequestsCount() > 0 && serverInfo().getServedRequestsCount() == 0) {
				DEBUGLOG("PANIC: server closed connection without serving a single request");
				finishSession();
			}
		}
	}
	m_data_for_server.erase();
	list<HttpClientRequest *>::iterator iter = m_requests_sent_to_server.begin();
	for (; iter != m_requests_sent_to_server.end(); ++iter) {
		if (!(*iter)->isIdempotent()) {
			DEBUGLOG("One of the pending requests is not idempotent and can't be retried => giving up");
			finishSession();
		}
	}
	m_requests_sent_to_server.clear();
	if (!m_requests_to_be_served.empty())
		requestStream().dispatchMoreRequests();
}

bool
BFilter::isRequestSentToServer(const HttpClientRequest *request) const
{
	return (std::find(m_requests_sent_to_server.begin(),
		m_requests_sent_to_server.end(), request) != m_requests_sent_to_server.end());
}

