/*
    BFilter - a smart ad-filtering web proxy
    Copyright (C) 2002-2003  Joseph Artsimovich <joseph_a@mail.ru>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <cstring>
#include "crc32.h"
#include "gzipdecompressor.h"

GzipDecompressor::GzipDecompressor()
:	m_error(false), m_eof(false), m_stream_ready(false), m_state(STATE_HEAD)
{}

GzipDecompressor::~GzipDecompressor()
{
	if (m_stream_ready)
		inflateEnd(&m_strm);
}

void
GzipDecompressor::reset()
{
	if (m_stream_ready) {
		inflateEnd(&m_strm);
		m_stream_ready = false;
	}
	m_data.erase();
	m_error = false;
	m_eof = false;
	m_state = STATE_HEAD;
}

bool
GzipDecompressor::readOutput(string& output)
{
	output.erase();
	if (m_error)
		return false;
	if (m_data.empty())
		return true;
	bool res = doReadOutput(output);
	m_error = !res || (m_eof && !m_data.empty());
	return res;
}

bool
GzipDecompressor::doReadOutput(string& output)
{
	while (true) {
		switch (m_state) {
		case STATE_HEAD:
			if (m_data.length() < 10)
				return true;
			if ((unsigned char)m_data[0] != 0x1f || (unsigned char)m_data[1] != 0x8b)
				return false;
			if (m_data[2] != 8) // wrong compression method
				return false;
			m_flg = m_data[3];
			if (m_flg & 0xe0) // reserved flags are not clear
				return false;
			m_head_crc32 = CRC32::compute(m_data.substr(0, 10));
			m_data.erase(0, 10);
			m_state = STATE_FEXTRA;
			break;
		case STATE_FEXTRA:
			if (m_flg & (1<<2)) {
				if (m_data.length() < 2)
					return true;
				string::size_type xlen = (string::size_type)m_data[0]+((string::size_type)m_data[1]<<8);
				if (m_data.length() < 2+xlen)
					return true;
				m_head_crc32 = CRC32::update(m_data.substr(0, 2+xlen), m_head_crc32);
				m_data.erase(0, 2+xlen);
			}
			m_state = STATE_FNAME;
			break;
		case STATE_FNAME:
			if (m_flg & (1<<3)) {
				string::size_type pos = m_data.find('\0');
				if (pos == string::npos)
					return true;
				m_head_crc32 = CRC32::update(m_data.substr(0, pos+1), m_head_crc32);
				m_data.erase(0, pos+1);
			}
			m_state = STATE_FCOMMENT;
			break;
		case STATE_FCOMMENT:
			if (m_flg & (1<<4)) {
				string::size_type pos = m_data.find('\0');
				if (pos == string::npos)
					return true;
				m_head_crc32 = CRC32::update(m_data.substr(0, pos+1), m_head_crc32);
				m_data.erase(0, pos+1);
			}
			m_state = STATE_FHCRC;
			break;
		case STATE_FHCRC:
			if (m_flg & (1<<1)) {
				if (m_data.length() < 2)
					return true;
				unsigned char b1 = m_data[0], b2 = m_data[1];
				unsigned long crc16 = (unsigned long)b1+((unsigned long)b2<<8);
				if (crc16 != m_head_crc32 & 0x0000FFFFL)
					return false;
				m_data.erase(0, 2);
			}
			m_state = STATE_DATA;
			m_data_length = 0L;
			m_data_crc32 = 0L;
			break;
		case STATE_DATA:
			if (m_data.empty())
				return true;
			if (!m_stream_ready) {
				m_strm.zalloc = Z_NULL;
				m_strm.zfree = Z_NULL;
				m_strm.opaque = Z_NULL;
				m_strm.next_in = Z_NULL;
				inflateInit2(&m_strm, -MAX_WBITS);
				m_stream_ready = true;
			}
			int res;
			do {
				m_strm.next_in = m_inbuf;
				string::size_type indata_size = sizeof(m_inbuf)<m_data.length() ? sizeof(m_inbuf) : m_data.length();
				m_strm.avail_in = indata_size;
				memcpy(m_inbuf, m_data.c_str(), indata_size);
				do {
					m_strm.next_out = m_outbuf;
					m_strm.avail_out = sizeof(m_outbuf);
					res = inflate(&m_strm, Z_SYNC_FLUSH);
					string dec_data(m_outbuf, m_outbuf+sizeof(m_outbuf)-m_strm.avail_out);
					m_data_crc32 = CRC32::update(dec_data, m_data_crc32);
					m_data_length += dec_data.length();
					output += dec_data;
				} while (res == Z_OK && m_strm.avail_out == 0);
				m_data.erase(0, indata_size-m_strm.avail_in);
			} while (res == Z_OK && !m_data.empty());
			if (res == Z_STREAM_END) {
				if (m_stream_ready) {
					inflateEnd(&m_strm);
					m_stream_ready = false;
				}
				m_state = STATE_CRC32_ISIZE;
				break;
			} else if (m_eof) {
				return false;
			}
			return (res == Z_OK);
			break;
		case STATE_CRC32_ISIZE:
			if (m_data.length() < 8)
				return true;
			unsigned long values[8];
			for (int i = 0; i < 8; i++) {
				values[i] = (unsigned char)m_data[i];
			}
			m_data.erase(0, 8);
			// stupid compiler - if I ommit the parentheses around the
			// blabla|blabla|... expression, the result is wrong
			if (m_data_crc32 != (values[0]|(values[1]<<8)|(values[2]<<16)|(values[3]<<24))) {
				return false;
			}
			if (m_data_length != (values[4]|(values[5]<<8)|(values[6]<<16)|(values[7]<<24))) {
				return false;
			}
			m_state = STATE_HEAD;
			break;
		default:
			return false;
		}
	}
	return false;
}

