/*
    BFilter - a smart ad-filtering web proxy
    Copyright (C) 2002-2003  Joseph Artsimovich <joseph_a@mail.ru>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "htmlstack.h"

HTMLStack::HTMLStack()
:	m_top(0), m_bottom(0)
{}

HTMLStack::HTMLStack(const HTMLStack& stack)
:	m_top(0), m_bottom(0)
{
	copyInClearEnvironment(stack);
}

HTMLStack&
HTMLStack::operator=(const HTMLStack& stack)
{
	clear();
	copyInClearEnvironment(stack);
	return *this;
}

void
HTMLStack::copyInClearEnvironment(const HTMLStack& stack)
{
	map<HTMLNode *, HTMLNode *> trans_old_new;
	map<HTMLNode *, HTMLNode *> trans_new_old;
	trans_old_new[0] = 0;
	trans_new_old[0] = 0;
	set<HTMLNode *>::const_iterator iter = stack.m_nodes.begin();
	for (; iter != stack.m_nodes.end(); ++iter) {
		HTMLNode *new_node = (*iter)->clone();
		m_nodes.insert(new_node);
		trans_old_new[*iter] = new_node;
		trans_new_old[new_node] = *iter;
	}
	iter = m_nodes.begin();
	for (; iter != m_nodes.end(); ++iter) {
		HTMLNode *new_node = *iter;
		HTMLNode *old_node = trans_new_old[new_node];
		new_node->m_parent = trans_old_new[old_node->m_parent];
		list<HTMLNode *>::iterator iter2 = new_node->m_children.begin();
		for (; iter2 != new_node->m_children.end(); ++iter2) {
			*iter2 = trans_old_new[*iter2];
		}
	}
	m_top = (HTMLElementNode *)trans_old_new[stack.m_top];
	m_bottom = (HTMLElementNode *)trans_old_new[stack.m_bottom];
}

HTMLStack::~HTMLStack()
{
	set<HTMLNode *>::iterator iter = m_nodes.begin();
	for (; iter != m_nodes.end(); ++iter) {
		delete *iter;
	}
}

void
HTMLStack::appendElement(const HTMLElementNode& el)
{
	if (!m_top)
		return;
	HTMLElementNode *new_el = new HTMLElementNode(el);
	m_nodes.insert(new_el);
	m_top->appendChild(new_el);
}

void
HTMLStack::appendText(const string& text)
{
	if (!m_top)
		return;
	HTMLTextNode *new_el = new HTMLTextNode(text);
	m_nodes.insert(new_el);
	m_top->appendChild(new_el);
}

void
HTMLStack::clearTop()
{
	if (!m_top)
		return;
	list<HTMLNode *> children = m_top->getChildren();
	m_top->removeAllChildren();
	list<HTMLNode *>::iterator iter = children.begin();
	for (; iter != children.end(); ++iter) {
		HTMLNode *node = *iter;
		m_nodes.erase(node);
		delete node;
	}
}

void
HTMLStack::pushElement(const HTMLElementNode& el)
{
	HTMLElementNode *new_el = new HTMLElementNode(el);
	m_nodes.insert(new_el);
	if (m_top) {
		m_top->appendChild(new_el);
		m_top = new_el;
	} else {
		new_el->m_parent = 0;
		m_top = m_bottom = new_el;
	}
}

void
HTMLStack::popElement()
{
	if (!m_top)
		return;
	clearTop();
	HTMLElementNode *parent = (HTMLElementNode *)m_top->getParent();
	if (parent)
		parent->removeChild(m_top);
	m_nodes.erase(m_top);
	delete m_top;
	m_top = parent;
	if (!m_top)
		m_bottom = 0;
}

void
HTMLStack::walkOn()
{
	if (!m_top)
		return;
		
	if (m_top->m_parent) {
		m_top = (HTMLElementNode *)m_top->m_parent;
	} else {
		popElement();
	}
}

void
HTMLStack::trimTo(const HTMLElementNode *el)
{
	while (m_top && m_top != el) {
		popElement();
	}
	clearTop();
}

void
HTMLStack::clear()
{
	set<HTMLNode *>::iterator iter = m_nodes.begin();
	for (; iter != m_nodes.end(); ++iter) {
		delete *iter;
	}
	m_nodes.clear();
	m_top = m_bottom = 0;
}

string
HTMLStack::toString(bool xhtml) const
{
	if (xhtml) {
		return m_bottom ? m_bottom->toString(xhtml) : "";
	} else {
		if (!m_bottom)
			return "";
		set<const HTMLNode *> noclose_nodes;
		HTMLNode *node = m_top;
		while (node) {
			noclose_nodes.insert(node);
			node = node->m_parent;
		}
		return m_bottom->toString(xhtml, &noclose_nodes);
	}
}

