/*
    BFilter - a smart ad-filtering web proxy
    Copyright (C) 2002-2003  Joseph Artsimovich <joseph_a@mail.ru>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "httpheaderelement.h"
#include <cstring>

const char *HttpHeaderElement::m_char_safety_table = HttpHeaderElement::prepareCharSafetyTable();
const char *HttpHeaderElement::m_char_escape_table = HttpHeaderElement::prepareCharEscapeTable();

bool
HttpHeaderElement::hasParameter(const string& pname) const
{
	list<NameValuePair>::const_iterator iter = m_params.begin();
	for (; iter != m_params.end(); ++iter) {
		if (iter->getName() == pname) {
			return true;
		}
	}
	return false;
}

NameValuePair
HttpHeaderElement::getParameterByName(const string& pname) const
{
	list<NameValuePair>::const_iterator iter = m_params.begin();
	for (; iter != m_params.end(); ++iter) {
		if (iter->getName() == pname) {
			return *iter;
		}
	}
	return NameValuePair("", "");
}

list<NameValuePair>
HttpHeaderElement::getParametersByName(const string& pname) const
{
	list<NameValuePair> res;
	list<NameValuePair>::const_iterator iter = m_params.begin();
	for (; iter != m_params.end(); ++iter) {
		if (iter->getName() == pname) {
			res.push_back(*iter);
		}
	}
	return res;
}

string
HttpHeaderElement::toString() const
{
	string res;
	res += getName();
	if (!getValue().empty()) {
		res += "="+valueToString(getValue());
	}
	list<NameValuePair>::const_iterator iter = m_params.begin();
	for (; iter != m_params.end(); ++iter) {
		res += ";"+iter->getName()+"="+valueToString(iter->getValue());
	}
	return res;
}

string
HttpHeaderElement::valueToString(const string& value)
{
	return valueRequiresQuoting(value) ? quoteValue(value) : value;
}

string
HttpHeaderElement::quoteValue(const string& value)
{
	string res = "\"";
	const char *data = value.data();
	string::size_type length = value.length();
	for (string::size_type i = 0; i < length; i++) {
		unsigned char ch = (unsigned char)data[i];
		if (ch == '"' || ch == '\\' || m_char_escape_table[ch] != ch) {
			res.append(1, '\\');
			res.append(1, m_char_escape_table[ch]);
		} else {
			res.append(1, ch);
		}
	}
	res += "\"";
	return res;
}

bool
HttpHeaderElement::valueRequiresQuoting(const string& value)
{
	const char *data = value.data();
	string::size_type length = value.length();
	for (string::size_type i = 0; i < length; i++) {
		if (!m_char_safety_table[(unsigned char)data[i]]) {
			return true;
		}
	}
	return false;
}

const char *
HttpHeaderElement::prepareCharSafetyTable()
{
	static char table[256];
	memset(table, 0, sizeof(table));
	memset(table+32, 1, sizeof(table[0])*(128-32));
	static const unsigned char more_unsafe[] = " <>\"{}|\\^~[]`#%\177";
	for (unsigned int i = 0; i < sizeof(more_unsafe); i++) {
		table[more_unsafe[i]] = 0;
	}
	return table;
}

const char *
HttpHeaderElement::prepareCharEscapeTable()
{
	static char table[256];
	for (unsigned int i = 0; i < 256; i++) {
		table[i] = (unsigned char)i;
	}
	table['\a'] = 'a';
	table['\b'] = 'b';
	table['\f'] = 'f';
	table['\n'] = 'n';
	table['\r'] = 'r';
	table['\t'] = 't';
	table['\v'] = 'v';
	return table;
}

