/*
    BFilter - a smart ad-filtering web proxy
    Copyright (C) 2002-2003  Joseph Artsimovich <joseph_a@mail.ru>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "responsefilterchain.h"

ResponseFilterChain::ResponseFilterChain(const HttpResponseMetadata& metadata, AbstractResponseFilter *first_filter)
:	m_metadata(metadata), m_waiting_for(m_filters.end())
{
	m_filters.push_back(first_filter);
	first_filter->consumeMetadata(m_metadata);
	if (!first_filter->isMetadataReady()) {
		--m_waiting_for;
	}
}

ResponseFilterChain::~ResponseFilterChain()
{
	list<AbstractResponseFilter *>::iterator iter = m_filters.begin();
	for (; iter != m_filters.end(); ++iter) {
		delete *iter;
	}
}

void
ResponseFilterChain::addFilter(AbstractResponseFilter *filter)
{
	m_filters.push_back(filter);
	if (m_waiting_for == m_filters.end()) {
		filter->consumeMetadata(m_metadata);
		if (!filter->isMetadataReady()) {
			--m_waiting_for;
		}
	}
}

bool
ResponseFilterChain::isMetadataReady() const
{
	// gcc 2.x has a problem with m_waiting_for == m_filters.end()
	typedef list<AbstractResponseFilter *>::const_iterator cit_type;
	return ((cit_type)m_waiting_for == (cit_type)m_filters.end());
}

void
ResponseFilterChain::consumeDataChunk(const string& data)
{
	if (data.empty()) {
		return;
	}
	string buf;
	list<AbstractResponseFilter *>::iterator first = m_filters.begin();
	(*first)->consumeDataChunk(data);
	if (m_waiting_for == first) {
		if ((*m_waiting_for)->isMetadataReady()) {
			if (++m_waiting_for != m_filters.end()) {
				(*m_waiting_for)->consumeMetadata(m_metadata);
			}
		} else {
			return;
		}
	}
	if (!(*first)->hasDataAvailable()) {
		return;
	}
	list<AbstractResponseFilter *>::iterator cur = first, prev = first;
	++cur;
	for (; cur != m_filters.end(); ++cur, ++prev) {
		(*prev)->readDataChunk(buf);
		(*cur)->consumeDataChunk(buf);
		if (m_waiting_for == cur) {
			if ((*m_waiting_for)->isMetadataReady()) {
				if (++m_waiting_for != m_filters.end()) {
					(*m_waiting_for)->consumeMetadata(m_metadata);
				}
			} else {
				return;
			}
		}
		if (!(*cur)->hasDataAvailable()) {
			return;
		}
	}
}

void
ResponseFilterChain::consumeEOF()
{
	string buf;
	list<AbstractResponseFilter *>::iterator first = m_filters.begin();
	(*first)->consumeEOF();
	if (m_waiting_for == first) {
		if ((*m_waiting_for)->isMetadataReady()) {
			if (++m_waiting_for != m_filters.end()) {
				(*m_waiting_for)->consumeMetadata(m_metadata);
			}
		} else {
			// the filter is not allowing the response to go out
			return;
		}
	}
	list<AbstractResponseFilter *>::iterator cur = first, prev = first;
	++cur;
	for (; cur != m_filters.end(); ++cur, ++prev) {
		(*prev)->readDataChunk(buf); // additional data may become available as a result of EOF
		(*cur)->consumeDataChunk(buf);
		(*cur)->consumeEOF();
		if (m_waiting_for == cur) {
			if ((*m_waiting_for)->isMetadataReady()) {
				if (++m_waiting_for != m_filters.end()) {
					(*m_waiting_for)->consumeMetadata(m_metadata);
				}
			} else {
				// the filter is not allowing the response to go out
				return;
			}
		}
	}
}

bool
ResponseFilterChain::hasDataAvailable() const
{
	return m_filters.back()->hasDataAvailable();
}

void
ResponseFilterChain::readDataChunk(string& collector)
{
	m_filters.back()->readDataChunk(collector);
}

bool
ResponseFilterChain::isEOF() const
{
	return (isMetadataReady() && m_filters.back()->isEOF());
}

