/*
    BFilter - a smart ad-filtering web proxy
    Copyright (C) 2002-2003  Joseph Artsimovich <joseph_a@mail.ru>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifndef RULES_H_
#define RULES_H_

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <string>
#include <map>
#include <list>
#include <sys/types.h>
#include <regex.h>
#include <conflistener.h>

using namespace std;

class Rules : protected ConfListener
{
public:
	void readRules();
	static void setConfDir(const string& conf_dir) { m_conf_dir = conf_dir; }
	static Rules& getInstance() { return m_instance; }
	int getFilterParam(const string& url) const;
	int getAdParam(const string& url) const;
	int getScriptsParam(const string& url) const;
	int getJSengineParam(const string& url) const;
	int getTargetBlankParam(const string& url) const;
protected:
	virtual ~Rules() {}
	virtual void processSection(const string& section, int line);
	virtual void processParameter(const string& section, const string& key, const string& value, int line);
	virtual void processError(const string& errmsg, int line);
private:
	class Rule
	{
	public:
		int m_filter;
		int m_ad;
		int m_scripts;
		int m_jsengine;
		int m_target_blank;
	};
	class Regex
	{
	public:
		Regex() : m_regex(0), m_refs(new int(1)) {}
		Regex(const string& regex, int flags) : m_refs(new int(1))
		{
			m_regex = new regex_t();
			if (regcomp(m_regex, regex.c_str(), flags)) {
				delete m_regex;
				m_regex = 0;
			}
		}
		Regex(const Regex& regex) : m_regex(regex.m_regex)
		{
			++*regex.m_refs;
			m_refs = regex.m_refs;
		}
		~Regex()
		{
			if (--*m_refs == 0) {
				if (m_regex) {
					regfree(m_regex);
					delete m_regex;
				}
				delete m_refs;
			}
		}
		Regex& operator=(const Regex& regex)
		{
			if (this == &regex)
				return *this;
			if (--*m_refs == 0) {
				if (m_regex) {
					regfree(m_regex);
					delete m_regex;
				}
				delete m_refs;
			}
			m_regex = regex.m_regex;
			++*regex.m_refs;
			m_refs = regex.m_refs;
			return *this;
		}
		bool operator==(const Regex& regex)
		{
			return (m_regex==regex.m_regex);
		}
		bool isNull() const { return !m_regex; }
		bool matches(const string& str) const
		{
			if (isNull())
				return false;
			return !regexec(m_regex, str.c_str(), 0, 0, 0);
		}
	private:
		regex_t *m_regex;
		mutable int *m_refs;
	};
	typedef map<string, map<string, int> > rule_storage_type;
	typedef map<int, Regex> grouped_regex_type;
	Rules();
	void warning(const string& msg, int line);
	int validateParameter(const string& key, const string& value, int line);
	void processDefaultRule(Rule& result, const Rule& base, rule_storage_type& rules);
	void groupRules();
	void groupBy(const string& key, const map<int, list<string> >& src, grouped_regex_type& dst);
	static string m_conf_dir;
	static Rules m_instance;
	grouped_regex_type m_global_re_by_filter;
	grouped_regex_type m_local_re_by_filter;
	grouped_regex_type m_global_re_by_ad;
	grouped_regex_type m_local_re_by_ad;
	grouped_regex_type m_global_re_by_scripts;
	grouped_regex_type m_local_re_by_scripts;
	grouped_regex_type m_global_re_by_jsengine;
	grouped_regex_type m_local_re_by_jsengine;
	grouped_regex_type m_global_re_by_target_blank;
	grouped_regex_type m_local_re_by_target_blank;
	Rule m_default_rule;
	Rule m_default_global_rule;
	Rule m_default_local_rule;
	
	// temporary data
	bool m_local;
	rule_storage_type m_global_rules;
	rule_storage_type m_local_rules;
};

#endif

