/*
    BFilter - a smart ad-filtering web proxy
    Copyright (C) 2002-2003  Joseph Artsimovich <joseph_a@mail.ru>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include "serverresponsestream.h"

#include <algorithm>
#include <sstream>
#include "bfilter.h"
#include "httperrormessages.h"
#include "state.h"
#include "conf.h"
#include "rules.h"
#include "network.h"
#include "acceptencoding.h"
#include "htmlresponsefilter.h"
#include "imageresponsefilter.h"
#include "compressorresponsefilter.h"
#include "decompressorresponsefilter.h"
#include "blackholeresponsefilter.h"
#include "httpresponsemetadata.h"
#include "httpstatusline.h"
#include "httpversion.h"
#include "replacementimage.h"
#include "utils.h"
#include "debug.h"

ServerResponseStream::~ServerResponseStream()
{
	if (m_filter_chain) {
		delete m_filter_chain;
	}
	if (m_stream_writer) {
		delete m_stream_writer;
	}
}

bool
ServerResponseStream::isResponseOnHeadRequest() const
{
	return (!m_owner.requestsSentToServer().empty() &&
	        m_owner.requestsSentToServer().front()->metadata().requestLine().getMethod() == "HEAD");
}

void
ServerResponseStream::eventError(int ecode, const char *info)
{
	if (!info) {
		info = "";
	}
	DEBUGLOG("ServerResponseStream::eventError(ecode = " << ecode << ")");
	
	const HttpClientRequest* request = m_owner.requestsSentToServer().empty() ? 0 : m_owner.requestsSentToServer().front();
	if (!request) {
		m_owner.finishSession();
	}
	
	switch (ecode) {
	case ERR_PARSING_HEADER:
		if (!isForwardingInProgress()) {
			m_owner.dataForClient(HttpErrorMessages::errMalformedResponseHeader(request->getOrigRequestLine()));
		}
		m_owner.finishSession();
		break;
	case ERR_PARSING_CHUNK_HEADER:
		if (!isForwardingInProgress()) {
			m_owner.dataForClient(HttpErrorMessages::errMalformedChunkInResponse(request->getOrigRequestLine()));
		}
		m_owner.finishSession();
		break;
	case ERR_UNEXPECTED_EOF:
		if (!isForwardingInProgress()) {
			m_owner.dataForClient(HttpErrorMessages::errUnexpectedServerEOF(request->getOrigRequestLine()));
		}
		m_owner.finishSession();
		break;
	default:
		if (!isForwardingInProgress()) {
			m_owner.dataForClient(HttpErrorMessages::errInternal(request->getOrigRequestLine(),
				"Unhandled error in ServerResponseStream: "+num2str(ecode)));
		}
		m_owner.finishSession();
	}
}

void
ServerResponseStream::eventGotMetadata(const HttpResponseMetadata& orig_metadata)
{
	HttpResponseMetadata metadata(orig_metadata);

	if (m_owner.requestsSentToServer().empty()) {
		DEBUGLOG2("PANIC: response to a non existent request");
		m_owner.finishSession();
	}
	
	HttpClientRequest *active_request = m_owner.requestsSentToServer().front();
	
	DEBUGLOG("Response begins (active request: " << active_request->metadata().requestLine().getURI().toString() << ")");
	
	if (m_filter_chain) {
		DEBUGLOG2("BUG");
		delete m_filter_chain;
		m_filter_chain = 0;
	}
	if (m_stream_writer) {
		DEBUGLOG2("BUG");
		delete m_stream_writer;
		m_stream_writer = 0;
	}
	
	if (metadata.statusLine().getHttpVersion() > m_owner.serverInfo().getProtocolCompability()) {
		m_owner.serverInfo().setProtocolCompability(metadata.statusLine().getHttpVersion());
	}
	
	prepareResponseForForwarding(metadata, active_request);
	
	// Some servers set Content-Encoding: gzip for .tar.gz files.
	// This is very wrong: it indicates to the user agent that it has to decompress the content.
	// As a result, you might get a tar file with the .tar.gz extension.
	if (metadata.headers().getHeader("Content-Encoding").getValue() == "gzip") {
		if (metadata.headers().getHeader("Content-Type").getValue() == "application/x-tar") {
			metadata.headers().removeHeader("Content-Encoding");
			metadata.headers().setHeader(HttpHeader("Content-Type", "application/x-tgz"));
		} else if (metadata.headers().getHeader("Content-Type").getValue() == "application/x-gzip") {
			metadata.headers().removeHeader("Content-Encoding");
			metadata.headers().setHeader(HttpHeader("Content-Type", "application/octet-stream"));
		} else {
			string req_path = strtolower(active_request->metadata().requestLine().getURI().getDecodedPath());
			string::size_type tgz_pos = req_path.rfind(".tgz");
			string::size_type tar_gz_pos = req_path.rfind(".tar.gz");
			if ((tgz_pos != string::npos && tgz_pos == req_path.length()-4)
			    || (tar_gz_pos != string::npos && tar_gz_pos == req_path.length()-7)) {
				metadata.headers().removeHeader("Content-Encoding");
				metadata.headers().setHeader(HttpHeader("Content-Type", "application/x-tgz"));
			} else {
				string::size_type gz_pos = req_path.rfind(".gz");
				if (gz_pos != string::npos && gz_pos == req_path.length()-3) {
					metadata.headers().removeHeader("Content-Encoding");
					metadata.headers().setHeader(HttpHeader("Content-Type", "application/octet-stream"));
				}
			}
		}
	}
	
	const string content_enc = metadata.headers().getHeader("Content-Encoding").getValue();
	const string content_type = metadata.headers().getHeader("Content-Type").getValue();
	AcceptEncoding accept_enc(active_request->metadata().headers().getHeader("Accept-Encoding"));
	int filter_rule = Rules::getInstance().getFilterParam(active_request->getAbsoluteURI().toString());
	int status_code = metadata.statusLine().getCode();
	m_response_status_code = status_code;
	bool has_body = !(metadata.getBodyStatus() == HttpMessageMetadata::NO_BODY_BUT_PRESERVE_HEADERS ||
		(metadata.getBodyStatus() == HttpMessageMetadata::BODY_SIZE_KNOWN && metadata.getBodySize() == 0));
	bool no_transform = active_request->metadata().headers().getHeader("Cache-Control").elements().hasElement("no-transform");
	
	if (status_code >= 100 && status_code < 200 &&
	    active_request->getOrigRequestLine().getHttpVersion() < HttpVersion::HTTP_1_1) {
		DEBUGLOG("1xx response to a < HTTP/1.1 request => applying BlackHoleResponseFilter");
		addFilterToChain(metadata, new BlackHoleResponseFilter);
		return;
	}
	
	bool need_filter_image = (status_code != 304 && !(status_code >= 100 && status_code < 200)
		&& has_body && !no_transform
		&& active_request->getInterceptHandler() == HttpClientRequest::IH_IMAGE_ANALYZE);
	
	bool need_filter_html = (!need_filter_image && (status_code == 200 || status_code >= 400)
		&& has_body && !no_transform && filter_rule != 0
		&& content_type.find("text/html") == 0 && ::State::getInstance().getFiltering());
	
	// this must go before setting up filters - it modifies metadata
	if (active_request->getInterceptHandler() == HttpClientRequest::IH_IMAGE_ANALYZE
	    && (status_code >= 301 && status_code <= 303) || status_code == 307) {
		URI redirect(metadata.headers().getHeader("Location").getValue());
		if (redirect.isAbsolute()) {
			URI subst_uri(redirect);
			subst_uri.removeFragment();
			subst_uri.setAbsoluteRawPath("bf.intercept-request/image_analyze");
			ostringstream strm;
			strm << "url=" << URI::escape(redirect.toString())
				<< "&w=" << active_request->getInterceptParam("w")
				<< "&h=" << active_request->getInterceptParam("h")
				<< "&iar=" << active_request->getInterceptParam("iar")
				<< "&ea=" << active_request->getInterceptParam("ea")
				<< "&ei=" << active_request->getInterceptParam("ei")
				<< "&sa=" << active_request->getInterceptParam("sa")
				<< "&si=" << active_request->getInterceptParam("si")
				<< "&target=" << URI::escape(active_request->getInterceptParam("target"))
				<< "&redir=1&cookie="
				<< ((metadata.headers().hasHeader("Set-Cookie") || metadata.headers().hasHeader("Set-Cookie2")) ? "1" : "0");
			subst_uri.setRawQuery(strm.str());
			metadata.headers().setHeader(HttpHeader("Location", subst_uri.toString()));
		} else {
			// TODO
		}
	}
	
	list<HttpHeaderElement> enc_elements;
	if (metadata.headers().hasHeader("Content-Encoding")) {
		const list<HttpHeaderElement>& elements = metadata.headers().getHeaderPtr("Content-Encoding")->elements().getElements();
		enc_elements.insert(enc_elements.end(), elements.begin(), elements.end());
	}
	if (metadata.headers().hasHeader("Transfer-Encoding")) {
		const list<HttpHeaderElement>& elements = metadata.headers().getHeaderPtr("Transfer-Encoding")->elements().getElements();
		enc_elements.insert(enc_elements.end(), elements.begin(), elements.end());
	}
	list<HttpHeaderElement>::iterator iter = enc_elements.begin();
	bool unknown_encoding = false;
	for (; iter != enc_elements.end(); ++iter) {
		if (iter->getLowercaseName() != "gzip" && iter->getLowercaseName() != "deflate") {
			unknown_encoding = true;
			break;
		}
	}
	if (unknown_encoding) {
		need_filter_image = false;
		need_filter_html = false;
	} else {
		list<HttpHeaderElement>::reverse_iterator iter2 = enc_elements.rbegin();
		for (; iter2 != enc_elements.rend(); ++iter2) {
			AbstractResponseFilter *filter;
			if (iter2->getLowercaseName() == "gzip") {
				filter = new DecompressorResponseFilter(DecompressorResponseFilter::C_GZIP);
			} else {
				filter = new DecompressorResponseFilter(DecompressorResponseFilter::C_DEFLATE);
			}
			addFilterToChain(metadata, filter);
		}
	}
	
	if (need_filter_html) {
		addFilterToChain(metadata, new HTMLResponseFilter(m_owner.htmlProcessor()));
		m_owner.htmlProcessor().setBaseURL(active_request->getAbsoluteURI());
	}
	if (need_filter_image) {
		int ad_rule = str2num<int>(active_request->getInterceptParam("iar"));
		int width = str2num<int>(active_request->getInterceptParam("w"));
		int height = str2num<int>(active_request->getInterceptParam("h"));
		string target = active_request->getInterceptParam("target");
		bool external_link = (active_request->getInterceptParam("ea") == "1");
		bool external_img = (active_request->getInterceptParam("ei") == "1");
		bool suspicious_link = (active_request->getInterceptParam("sa") == "1");
		bool suspicious_img = (active_request->getInterceptParam("si") == "1");
		bool redir = (active_request->getInterceptParam("redir") == "1");
		bool cookie = (active_request->getInterceptParam("cookie") == "1");
		if (!cookie && metadata.headers().hasHeader("Set-Cookie") || metadata.headers().hasHeader("Set-Cookie2")) {
			cookie = true;
		}
		AbstractResponseFilter *filter = new ImageResponseFilter(ad_rule, width, height,
			target, external_link, external_img, suspicious_link, suspicious_img, redir, cookie);
		addFilterToChain(metadata, filter);
	}
	
	if (has_body && content_type.find("text/") == 0 && Conf::getInstance().clientCompression()) {
		list<string> possible_encodings;
		possible_encodings.push_back("gzip");
		possible_encodings.push_back("deflate");
		string best_encoding = accept_enc.bestAccepted(possible_encodings);
		if (best_encoding == "deflate") {
			AbstractResponseFilter *filter = new CompressorResponseFilter(CompressorResponseFilter::C_DEFLATE);
			addFilterToChain(metadata, filter);
		} else if (best_encoding == "gzip") {
			AbstractResponseFilter *filter = new CompressorResponseFilter(CompressorResponseFilter::C_GZIP);
			addFilterToChain(metadata, filter);
		}
	}

	if (!m_filter_chain) {
		m_stream_writer = new HttpStreamWriter(*this, metadata,
			active_request->getOrigRequestLine().getHttpVersion() >= HttpVersion::HTTP_1_1,
			active_request->isClosingRequest()
		);
	} else {
		metadata.headers().removeHeader("Content-Length");
	}
}

void
ServerResponseStream::eventGotContentPart(const string& data)
{
	if (!m_filter_chain) {
		m_stream_writer->write(data);
	} else {
		m_filter_chain->consumeDataChunk(data);
		if (!m_stream_writer && m_filter_chain->isMetadataReady()) {
			m_stream_writer = new HttpStreamWriter(*this, m_filter_chain->getMetadata(),
				m_owner.requestsSentToServer().front()->getOrigRequestLine().getHttpVersion()
				>= HttpVersion::HTTP_1_1, m_owner.requestsSentToServer().front()->isClosingRequest()
			);
		}
		if (m_filter_chain->hasDataAvailable() && m_stream_writer) {
			string out;
			m_filter_chain->readDataChunk(out);
			m_stream_writer->write(out);
		}
	}
}

void
ServerResponseStream::eventMessageEnds(bool connection_needs_closing)
{
	DEBUGLOG("ServerResponseStream::eventMessageEnds(conn_close = " << connection_needs_closing << ")");
	m_owner.serverInfo().registerResponse(m_response_status_code);
	if (connection_needs_closing) {
		m_owner.serverInfo().setPipeliningCapability(ServerInfo::P_NOT_OK);
	}
	
#ifdef DEBUG
	if (m_owner.requestsSentToServer().empty()) {
		DEBUGLOG2("BUG");
	}
#endif
	HttpClientRequest *active_request = m_owner.requestsSentToServer().front();
	
	bool suppress_response = false;
	if (m_filter_chain) {
		m_filter_chain->consumeEOF();
		if (!m_stream_writer) {
			if (m_filter_chain->isMetadataReady()) {
				m_stream_writer = new HttpStreamWriter(*this, m_filter_chain->getMetadata(),
					active_request->getOrigRequestLine().getHttpVersion() >= HttpVersion::HTTP_1_1,
					active_request->isClosingRequest()
				);
			} else {
				suppress_response = true;
			}
		}
		if (!suppress_response) {
			if (m_filter_chain->hasDataAvailable()) {
				string out;
				m_filter_chain->readDataChunk(out);
				m_stream_writer->write(out);
			}
			if (m_filter_chain->isEOF()) {
				m_stream_writer->end();
			}
		} else {
#ifdef DEBUG
			if (m_response_status_code >= 100 && m_response_status_code < 200) {
				DEBUGLOG("1xx response blocked");
			} else {
				DEBUGLOG("BUG");
			}
#endif
		}
	} else {
		m_stream_writer->end();
	}
	
	if (m_filter_chain) {
		delete m_filter_chain;
		m_filter_chain = 0;
	}
	if (m_stream_writer) {
		delete m_stream_writer;
		m_stream_writer = 0;
	}
	if (!suppress_response) {
		if (!active_request->isComplete()) {
			DEBUGLOG("request is not yet complete, but a full response has arrived [a rare case, but possible]");
			active_request->setAlreadyServed();
		} else {
			delete active_request;
		}
		m_owner.requestsSentToServer().pop_front();
		m_owner.requestsToBeServed().pop_front();
		DEBUGLOG("The request have been served");
		list<HttpClientRequest *>::iterator iter = m_owner.requestsToBeServed().begin();
		while (iter != m_owner.requestsToBeServed().end()) {
			HttpClientRequest *request = *iter;
			if (request->isToBeSent() || !request->isComplete()) {
				break;
			}
			processSubstitutionRequest(request);
			delete request;
			m_owner.requestsToBeServed().erase(iter++);
			DEBUGLOG("Substitution request served (requests_sent_to_server: " << m_owner.requestsSentToServer().size()
				<< ", requests_to_be_served: " << m_owner.requestsToBeServed().size() << ")");
		}

		if (m_owner.requestsSentToServer().empty()) {
			if (m_owner.serverInfo().getPipeliningCapability() == ServerInfo::P_TESTING) {
				m_owner.serverInfo().setPipeliningCapability(ServerInfo::P_OK);
				DEBUGLOG("The server has responded to a pipelined request - marking as P_OK");
			}
			DEBUGLOG("Response ends (requests_sent_to_server: " << m_owner.requestsSentToServer().size()
				<< ", requests_to_be_served: " << m_owner.requestsToBeServed().size() << ")");
			if (connection_needs_closing) {
				if (Network::isValidSocket(m_owner.m_ssock)) {
					m_owner.closeServerConnection(false);
				}
			}
			m_owner.requestStream().dispatchMoreRequests();
		}
	}
	
}

void
ServerResponseStream::moreData(const string& data)
{
	m_forwarding_in_progress = true;
	m_owner.dataForClient(data);
}

void
ServerResponseStream::done(bool conn_close)
{
	m_forwarding_in_progress = false;
	if (conn_close) {
		m_owner.finishSession();
	}
}

void
ServerResponseStream::prepareResponseForForwarding(HttpResponseMetadata& metadata, const HttpClientRequest *active_request)
{
	HttpHeader connection_hdr = metadata.headers().getHeader("Connection");
	connection_hdr.elements().appendElements(metadata.headers().getHeader("Proxy-Connection").elements());
	list<HttpHeaderElement> hop_by_hop_headers = connection_hdr.elements().getElements();
	list<HttpHeaderElement>::iterator iter = hop_by_hop_headers.begin();
	for (; iter != hop_by_hop_headers.end(); ++iter) {
		metadata.headers().removeHeader(iter->getName());
	}
	metadata.headers().removeHeader("Keep-Alive");
	metadata.headers().removeHeader("Connection");
	metadata.headers().removeHeader("Proxy-Connection");
	metadata.statusLine().setHttpVersion(
		std::min(active_request->getOrigRequestLine().getHttpVersion(), HttpVersion::HTTP_1_1)
	);
	if (metadata.statusLine().getHttpVersion() < HttpVersion::HTTP_1_1) {
		if (metadata.statusLine().getCode() == 303 || metadata.statusLine().getCode() == 307) {
			metadata.statusLine().setCodeAndMessage(302);
		}
	}
#ifdef DEBUG
	metadata.headers().setHeader(HttpHeader("X-in-response-to", active_request->getAbsoluteURI().toString()));
#endif
}

void
ServerResponseStream::addFilterToChain(HttpResponseMetadata& metadata, AbstractResponseFilter *filter)
{
	if (m_filter_chain) {
		m_filter_chain->addFilter(filter);
	} else {
		m_filter_chain = new ResponseFilterChain(metadata, filter);
	}
}

void
ServerResponseStream::processSubstitutionRequest(const HttpClientRequest *request)
{
	DEBUGLOG("processSubstitutionRequest()");
	if (request->getInterceptHandler() == HttpClientRequest::IH_IMAGE_SUBSTITUTE) {
		performImageSubstitution(request);
	} else {
		performErrorSubstitution(request);
	}
}

void
ServerResponseStream::performImageSubstitution(const HttpClientRequest *request)
{
	DEBUGLOG("performImageSubstitution()");
#ifdef DEBUG
	if (!request->isComplete()) {
		DEBUGLOG2("BUG");
	}
	if (request->isAlreadyServed()) {
		DEBUGLOG2("BUG");
	}
#endif
	bool closing_request = request->isClosingRequest();
	HttpVersion request_version = request->metadata().requestLine().getHttpVersion();
	HttpStatusLine status_line(std::min(request_version, HttpVersion::HTTP_1_1), 200);
	HttpResponseMetadata metadata(status_line, HttpHeadersCollection());
	HttpStreamWriter *stream_writer = 0;
	if (request->metadata().requestLine().getMethod() == "HEAD") {
		metadata.setBodyStatus(HttpMessageMetadata::NO_BODY_BUT_PRESERVE_HEADERS);
		metadata.headers().setHeader(HttpHeader("Content-Type", "image/gif"));
#ifdef DEBUG
		metadata.headers().setHeader(HttpHeader("X-in-response-to", request->metadata().requestLine().getURI().toString()));
#endif
		stream_writer = new HttpStreamWriter(*this, metadata, request_version>=HttpVersion::HTTP_1_1, closing_request);
		stream_writer->end();
	} else {
		int width = str2num<int>(request->getInterceptParam("w"));
		int height = str2num<int>(request->getInterceptParam("h"));
		ReplacementImage image(width, height, Conf::getInstance().drawBorders(), Conf::getInstance().getBorderColor());
		HttpHeadersCollection headers;
		metadata.setBodyStatus(HttpMessageMetadata::BODY_SIZE_KNOWN);
		metadata.setBodySize(image.getBinaryData().length());
		metadata.headers().setHeader(HttpHeader("Content-Type", image.getContentType()));
#ifdef DEBUG
		metadata.headers().setHeader(HttpHeader("X-in-response-to", request->metadata().requestLine().getURI().toString()));
#endif
		stream_writer = new HttpStreamWriter(*this, metadata, request_version>=HttpVersion::HTTP_1_1, closing_request);
		stream_writer->write(image.getBinaryData());
		stream_writer->end();
	}
	delete stream_writer;
}

void
ServerResponseStream::performErrorSubstitution(const HttpClientRequest *request)
{
	DEBUGLOG("performErrorSubstitution()");
	m_owner.dataForClient(request->getErrorResponse());
	m_owner.finishSession();
}

